@@ -5,13 +5,14 @@ Decorators
 from __future__ import annotations
 
 import ast
+import copy
 import inspect
 import logging
 import textwrap
 import warnings
 from functools import wraps
 from types import FunctionType
-from typing import Generator, Iterator, Optional, Protocol, Union, cast
+from typing import Iterator, Optional, Union, cast
 
 import __main__ as main
 
@@ -79,29 +80,70 @@ def verilogify(
     def generator_wrapper(*args, **kwargs):
         nonlocal context
         if kwargs:
-            raise RuntimeError(
+            warnings.warn(
                 "Keyword arguments not yet supported, use positional arguments only"
             )
+        for arg in args:
+            assert guard(arg, int)
+
         context.test_cases.append(args)
+
+        # Input inference
         if not context.input_types:
             context.input_types = [type(arg) for arg in args]
         else:
             context.check_input_types(args)
 
-        for result in func(*args, **kwargs):
-            if not isinstance(result, tuple):
-                result = (result,)
+        def tuplefy(either: Union[int, tuple[int]]) -> tuple[int]:
+            """
+            Converts int to tuple, otherwise returns input
+            """
+            if isinstance(either, int):
+                ret = (either,)
+            else:
+                ret = either
 
+            for value in ret:
+                try:
+                    assert guard(value, int)
+                except Exception as e:
+                    raise TypeError("Expected `int` type inputs and outputs") from e
+            return ret
+
+        # Always get output one-ahead of what func user sees
+        # For output type inference even if user doesn't use generator
+        instance = func(*args)
+        try:
+            result = cast(Union[int, tuple[int]], next(instance))
+            tupled_result = tuplefy(result)
             if not context.output_types:
                 logging.info(
-                    f"Using input `{result}` as reference for {func.__name__}'s I/O types"
+                    "Using input `%s` as reference for %s's I/O types",
+                    tupled_result,
+                    func.__name__,
                 )
-                context.output_types = [type(arg) for arg in result]
+                context.output_types = [type(arg) for arg in tupled_result]
                 context.default_output_vars()
             else:
-                context.check_output_types(result)
+                context.check_output_types(tupled_result)
+        except StopIteration:
+            pass
+        except Exception as e:
+            raise e
 
-        return func(*args, **kwargs)
+        @wraps(func)
+        def inside():
+            nonlocal result
+            for i, new_result in enumerate(instance):
+                tupled_new_result = tuplefy(new_result)
+                context.check_output_types(tupled_new_result)
+                yield result
+                result = new_result
+                if i > 10000:
+                    raise RuntimeError(f"{func.__name__} yields more than 10000 values")
+            yield result
+
+        return inside()
 
     @wraps(func)
     def function_wrapper(*_0, **_1):
@@ -121,9 +163,9 @@ def verilogify(
 
 def get_context(verilogified: FunctionType) -> ir.Context:
     """
-    Gets context from verilogified function
+    Gets a copy of the context from a verilogified function
     """
-    return verilogified._python2verilog_context  # type: ignore # pylint: disable=protected-access
+    return copy.deepcopy(verilogified._python2verilog_context)  # type: ignore # pylint: disable=protected-access
 
 
 def get_original_func(verilogified: FunctionType) -> FunctionType:
@@ -133,13 +175,23 @@ def get_original_func(verilogified: FunctionType) -> FunctionType:
     return verilogified._python2verilog_original_func  # type: ignore # pylint: disable=protected-access
 
 
-def get_expected(verilogified: FunctionType) -> Iterator[tuple[int, ...]]:
+def get_expected(
+    verilogified: FunctionType, max_yields_per_test_case: int = 10000
+) -> Iterator[tuple[int, ...]]:
     """
     Get expected output of testbench
+
+    Limits number of values generator can yield
     """
     generator_func = get_original_func(verilogified)
     for test in get_context(verilogified).test_cases:
-        yield from generator_func(*test)
+        logging.debug("Test case %s", test)
+        for i, value in enumerate(generator_func(*test)):
+            yield value
+            if i > max_yields_per_test_case:
+                raise RuntimeError(
+                    f"Generator yielded more than {max_yields_per_test_case} values"
+                )
 
 
 def get_actual_raw(
